"""A kinetic model of the methionine cycle.

See here for more about the methionine cycle:
https://doi.org/10.1021/acssynbio.3c00662

"""

from jax import numpy as jnp

from enzax.kinetic_model import (
    RateEquationModel,
    KineticModelStructure,
)
from enzax.rate_equations import (
    AllostericIrreversibleMichaelisMenten,
    Drain,
    IrreversibleMichaelisMenten,
    ReversibleMichaelisMenten,
)
from enzax.parameters import AllostericMichaelisMentenParameterSet

coords = {
    "enzyme": [
        "MAT1",
        "MAT3",
        "METH-Gen",
        "GNMT1",
        "AHC1",
        "MS1",
        "BHMT1",
        "CBS1",
        "MTHFR1",
        "PROT1",
    ],
    "drain": ["the_drain"],
    "rate": [
        "the_drain",
        "MAT1",
        "MAT3",
        "METH-Gen",
        "GNMT1",
        "AHC1",
        "MS1",
        "BHMT1",
        "CBS1",
        "MTHFR1",
        "PROT1",
    ],
    "metabolite": [
        "met-L",
        "atp",
        "pi",
        "ppi",
        "amet",
        "ahcys",
        "gly",
        "sarcs",
        "hcys-L",
        "adn",
        "thf",
        "5mthf",
        "mlthf",
        "glyb",
        "dmgly",
        "ser-L",
        "nadp",
        "nadph",
        "cyst-L",
    ],
    "km": [
        "met-L MAT1",
        "atp MAT1",
        "met-L MAT3",
        "atp MAT3",
        "amet METH-Gen",
        "amet GNMT1",
        "gly GNMT1",
        "ahcys AHC1",
        "hcys-L AHC1",
        "adn AHC1",
        "hcys-L MS1",
        "5mthf MS1",
        "hcys-L BHMT1",
        "glyb BHMT1",
        "hcys-L CBS1",
        "ser-L CBS1",
        "mlthf MTHFR1",
        "nadph MTHFR1",
        "met-L PROT1",
    ],
    "ki": [
        "MAT1",
        "METH-Gen",
        "GNMT1",
    ],
    "species": [
        "met-L",
        "atp",
        "pi",
        "ppi",
        "amet",
        "ahcys",
        "gly",
        "sarcs",
        "hcys-L",
        "adn",
        "thf",
        "5mthf",
        "mlthf",
        "glyb",
        "dmgly",
        "ser-L",
        "nadp",
        "nadph",
        "cyst-L",
    ],
    "balanced_species": [
        "met-L",
        "amet",
        "ahcys",
        "hcys-L",
        "5mthf",
    ],
    "unbalanced_species": [
        "atp",
        "pi",
        "ppi",
        "gly",
        "sarcs",
        "adn",
        "thf",
        "mlthf",
        "glyb",
        "dmgly",
        "ser-L",
        "nadp",
        "nadph",
        "cyst-L",
    ],
    "transfer_constant": [
        "METAT",
        "GNMT",
        "CBS",
        "MTHFR",
    ],
    "dissociation_constant": [
        "met-L MAT3",
        "amet MAT3",
        "amet GNMT1",
        "mlthf GNMT1",
        "amet CBS1",
        "amet MTHFR1",
        "ahcys MTHFR1",
    ],
}
dims = {
    "log_kcat": ["enzyme"],
    "log_enzyme": ["enzyme"],
    "log_drain": ["drain"],
    "log_km": ["km"],
    "dgf": ["metabolite"],
    "log_ki": ["ki"],
    "log_conc_unbalanced": ["unbalanced_species"],
    "log_transfer_constant": ["transfer_constant"],
    "log_dissociation_constant": ["dissociation_constant"],
}
parameters = AllostericMichaelisMentenParameterSet(
    log_kcat=jnp.log(
        jnp.array(
            [
                7.89577,  # MAT1
                19.9215,  # MAT3
                1.15777,  # METH-Gen
                10.5307,  # GNMT1
                234.284,  # AHC1
                1.77471,  # MS1
                13.7676,  # BHMT1
                7.02307,  # CBS1
                3.1654,  # MTHFR1
                0.264744,  # PROT1
            ]
        )
    ),
    log_enzyme=jnp.log(
        jnp.array(
            [
                0.000961712,  # MAT1
                0.00098812,  # MAT3
                0.00103396,  # METH-Gen
                0.000983692,  # GNMT1
                0.000977878,  # AHC1
                0.00105094,  # MS1
                0.000996603,  # BHMT1
                0.00134056,  # CBS1
                0.0010054,  # MTHFR1
                0.000995525,  # PROT1
            ]
        )
    ),
    log_drain=jnp.log(jnp.array([0.000850127])),
    dgf=jnp.array(
        [
            160.953,  # met-L
            -2263.31,  # atp
            -1055.95,  # pi
            -1943.8,  # ppi
            636.255,  # amet
            547.319,  # ahcys
            -161.373,  # gly
            -39.4573,  # sarcs
            44.2,  # hcys-L
            375.758,  # adn
            108.366,  # thf
            223.646,  # 5mthf
            198.009,  # mlthf
            173.094,  # glyb
            49.4547,  # dmgly
            -216.712,  # ser-L
            -2014.52,  # nadp
            -1948.58,  # nadph
            -46.4737,  # cyst-L
        ]
    ),
    log_km=jnp.log(
        jnp.array(
            [
                0.000106919,  # met-L MAT1
                0.00203015,  # atp MAT1
                0.00113258,  # met-L MAT3
                0.00236759,  # atp MAT3
                9.37e-06,  # amet METH-Gen
                0.000520015,  # amet GNMT1
                0.00253545,  # gly GNMT1
                2.32e-05,  # ahcys AHC1
                1.06e-05,  # hcys-L AHC1
                5.66e-06,  # adn AHC1
                1.71e-06,  # hcys-L MS1
                6.94e-05,  # 5mthf MS1
                1.98e-05,  # hcys-L BHMT1
                0.00845898,  # glyb BHMT1
                4.24e-05,  # hcys-L CBS1
                2.83e-06,  # ser-L CBS1
                8.08e-05,  # mlthf MTHFR1
                2.09e-05,  # nadph MTHFR1
                4.39e-05,  # met-L PROT1
            ]
        )
    ),
    log_ki=jnp.log(
        jnp.array(
            [
                0.000346704,  # MAT1
                5.56e-06,  # METH-Gen
                5.31e-05,  # GNMT1
            ]
        )
    ),
    log_conc_unbalanced=jnp.log(
        jnp.array(
            [
                # dataset1
                0.00131546,  # atp
                0.001,  # pi
                0.000500016,  # ppi
                0.00145177,  # gly
                1.00e-07,  # sarcs
                1.01e-06,  # adn
                2.24e-05,  # thf
                3.15e-06,  # mlthf
                0.00106758,  # glyb
                5.00e-05,  # dmgly
                0.0015873,  # ser-L
                1.22e-06,  # nadp
                0.000245139,  # nadph
                2.24e-06,  # cyst-L
            ]
        )
    ),
    temperature=jnp.array(298.15),
    log_transfer_constant=jnp.log(
        jnp.array(
            [
                0.107657,  # METAT
                131.207,  # GNMT
                1.03452,  # CBS
                0.392035,  # MTHFR
            ]
        )
    ),
    log_dissociation_constant=jnp.log(
        jnp.array(
            [
                0.00059999,  # met-L MAT3
                0.000316641,  # amet MAT3
                1.98e-05,  # amet GNMT1
                0.000228576,  # mlthf GNMT1
                9.30e-05,  # amet CBS1
                1.46e-05,  # amet MTHFR1
                2.45e-06,  # ahcys MTHFR1
            ]
        )
    ),
)
structure = KineticModelStructure(
    S=jnp.array(
        [
            [1, -1, -1, 0, 0, 0, 1, 1, 0, 0, -1],  # met-L b
            [0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0],  # atp
            [0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0],  # pi
            [0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0],  # ppi
            [0, 1, 1, -1, -1, 0, 0, 0, 0, 0, 0],  # amet b
            [0, 0, 0, 1, 1, -1, 0, 0, 0, 0, 0],  # ahcys b
            [0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 0],  # gly
            [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],  # sarcs
            [0, 0, 0, 0, 0, 1, -1, -1, -1, 0, 0],  # hcys-L b
            [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],  # adn
            [0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],  # thf
            [0, 0, 0, 0, 0, 0, -1, 0, 0, 1, 0],  # 5mthf b
            [0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0],  # mlthf
            [0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0],  # glyb
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],  # dmgly
            [0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0],  # ser-L
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],  # nadp
            [0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0],  # nadph
            [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],  # cyst-L
        ],
        dtype=jnp.float64,
    ),
    balanced_species=jnp.array(
        [
            0,  # met-L
            4,  # amet
            5,  # ahcys
            8,  # hcys-L
            11,  # 5mthf
        ]
    ),
    unbalanced_species=jnp.array(
        [
            1,  # atp
            2,  # pi
            3,  # ppi
            6,  # gly
            7,  #  sarcs
            9,  # adn
            10,  # thf
            12,  # mlthf
            13,  # glyb
            14,  # dmgly
            15,  # ser-L
            16,  # nadp
            17,  # nadph
            18,  # cyst-L
        ]
    ),
)
rate_equations = [
    Drain(sign=jnp.array(1.0), drain_ix=0),  # met-L source
    IrreversibleMichaelisMenten(  # MAT1
        kcat_ix=0,
        enzyme_ix=0,
        km_ix=jnp.array([0, 1], dtype=jnp.int16),
        ki_ix=jnp.array([0], dtype=jnp.int16),
        reactant_stoichiometry=jnp.array(
            [-1.0, -1.0, 1.0, 1.0, 1.0], dtype=jnp.int16
        ),
        ix_substrate=jnp.array([0, 1], dtype=jnp.int16),
        ix_ki_species=jnp.array([4], dtype=jnp.int16),
        substrate_km_positions=jnp.array([0, 1], dtype=jnp.int16),
        substrate_reactant_positions=jnp.array([0, 1], dtype=jnp.int16),
    ),
    AllostericIrreversibleMichaelisMenten(  # MAT3
        kcat_ix=1,
        enzyme_ix=1,
        km_ix=jnp.array([2, 3], dtype=jnp.int16),
        ki_ix=jnp.array([], dtype=jnp.int16),
        reactant_stoichiometry=jnp.array(
            [-1.0, -1.0, 1.0, 1.0, 1.0], dtype=jnp.int16
        ),
        ix_substrate=jnp.array([0, 1], dtype=jnp.int16),
        ix_ki_species=jnp.array([], dtype=jnp.int16),
        substrate_km_positions=jnp.array([0, 1], dtype=jnp.int16),
        substrate_reactant_positions=jnp.array([0, 1], dtype=jnp.int16),
        subunits=2,
        tc_ix=0,
        ix_dc_inhibition=jnp.array([], dtype=jnp.int16),
        ix_dc_activation=jnp.array([0, 1], dtype=jnp.int16),
        species_inhibition=jnp.array([], dtype=jnp.int16),
        species_activation=jnp.array([0, 4], dtype=jnp.int16),
    ),
    IrreversibleMichaelisMenten(  # METH
        kcat_ix=2,
        enzyme_ix=2,
        km_ix=jnp.array([4], dtype=jnp.int16),
        ki_ix=jnp.array([1], dtype=jnp.int16),
        reactant_stoichiometry=jnp.array([-1, 1], dtype=jnp.int16),
        ix_substrate=jnp.array([4], dtype=jnp.int16),
        ix_ki_species=jnp.array([5], dtype=jnp.int16),
        substrate_km_positions=jnp.array([0], dtype=jnp.int16),
        substrate_reactant_positions=jnp.array([0], dtype=jnp.int16),
    ),
    AllostericIrreversibleMichaelisMenten(  # GNMT1
        kcat_ix=3,
        enzyme_ix=3,
        km_ix=jnp.array([5, 6], dtype=jnp.int16),
        ki_ix=jnp.array([2], dtype=jnp.int16),
        reactant_stoichiometry=jnp.array(
            [-1.0, 1.0, -1.0, 1.0], dtype=jnp.int16
        ),
        ix_substrate=jnp.array([4, 6], dtype=jnp.int16),
        ix_ki_species=jnp.array([5], dtype=jnp.int16),
        substrate_km_positions=jnp.array([0, 1], dtype=jnp.int16),
        substrate_reactant_positions=jnp.array([0, 2], dtype=jnp.int16),
        subunits=4,
        tc_ix=1,
        ix_dc_activation=jnp.array([2], dtype=jnp.int16),
        ix_dc_inhibition=jnp.array([3], dtype=jnp.int16),
        species_inhibition=jnp.array([12], dtype=jnp.int16),
        species_activation=jnp.array([4], dtype=jnp.int16),
    ),
    ReversibleMichaelisMenten(  # AHC
        kcat_ix=4,
        enzyme_ix=4,
        km_ix=jnp.array([7, 8, 9], dtype=jnp.int16),
        ki_ix=jnp.array([], dtype=jnp.int16),
        reactant_stoichiometry=jnp.array([-1.0, 1.0, 1.0], dtype=jnp.int16),
        ix_substrate=jnp.array([5], dtype=jnp.int16),
        ix_product=jnp.array([8, 9], dtype=jnp.int16),
        ix_reactants=jnp.array([5, 8, 9], dtype=jnp.int16),
        ix_ki_species=jnp.array([], dtype=jnp.int16),
        substrate_km_positions=jnp.array([0], dtype=jnp.int16),
        product_km_positions=jnp.array([1, 2], dtype=jnp.int16),
        substrate_reactant_positions=jnp.array([0], dtype=jnp.int16),
        product_reactant_positions=jnp.array([1, 2], dtype=jnp.int16),
        water_stoichiometry=jnp.array(-1.0),
        reactant_to_dgf=jnp.array([5, 8, 9], dtype=jnp.int16),
    ),
    IrreversibleMichaelisMenten(  # MS
        kcat_ix=5,
        enzyme_ix=5,
        km_ix=jnp.array([10, 11], dtype=jnp.int16),
        ki_ix=jnp.array([], dtype=jnp.int16),
        reactant_stoichiometry=jnp.array([1, -1, 1, -1], dtype=jnp.int16),
        ix_substrate=jnp.array([8, 11], dtype=jnp.int16),
        ix_ki_species=jnp.array([], dtype=jnp.int16),
        substrate_km_positions=jnp.array([0, 1], dtype=jnp.int16),
        substrate_reactant_positions=jnp.array([1, 3], dtype=jnp.int16),
    ),
    IrreversibleMichaelisMenten(  # BHMT
        kcat_ix=6,
        enzyme_ix=6,
        km_ix=jnp.array([12, 13], dtype=jnp.int16),
        ki_ix=jnp.array([], dtype=jnp.int16),
        reactant_stoichiometry=jnp.array([1, -1, -1, 1], dtype=jnp.int16),
        ix_substrate=jnp.array([8, 13], dtype=jnp.int16),
        ix_ki_species=jnp.array([], dtype=jnp.int16),
        substrate_km_positions=jnp.array([0, 1], dtype=jnp.int16),
        substrate_reactant_positions=jnp.array([1, 2], dtype=jnp.int16),
    ),
    AllostericIrreversibleMichaelisMenten(  # CBS1
        kcat_ix=7,
        enzyme_ix=7,
        km_ix=jnp.array([14, 15], dtype=jnp.int16),
        ki_ix=jnp.array([], dtype=jnp.int16),
        reactant_stoichiometry=jnp.array([-1, -1, 1], dtype=jnp.int16),
        ix_substrate=jnp.array([8, 15], dtype=jnp.int16),
        ix_ki_species=jnp.array([], dtype=jnp.int16),
        substrate_km_positions=jnp.array([0, 1], dtype=jnp.int16),
        substrate_reactant_positions=jnp.array([0, 1], dtype=jnp.int16),
        subunits=2,
        tc_ix=2,
        ix_dc_activation=jnp.array([4], dtype=jnp.int16),
        ix_dc_inhibition=jnp.array([], dtype=jnp.int16),
        species_inhibition=jnp.array([4], dtype=jnp.int16),
        species_activation=jnp.array([], dtype=jnp.int16),
    ),
    AllostericIrreversibleMichaelisMenten(  # MTHFR
        kcat_ix=8,
        enzyme_ix=8,
        km_ix=jnp.array([16, 17], dtype=jnp.int16),
        ki_ix=jnp.array([], dtype=jnp.int16),
        reactant_stoichiometry=jnp.array([1, -1, 1, -1], dtype=jnp.int16),
        ix_substrate=jnp.array([12, 17], dtype=jnp.int16),
        ix_ki_species=jnp.array([], dtype=jnp.int16),
        substrate_km_positions=jnp.array([0, 1], dtype=jnp.int16),
        substrate_reactant_positions=jnp.array([1, 3], dtype=jnp.int16),
        subunits=2,
        tc_ix=3,
        ix_dc_activation=jnp.array([6], dtype=jnp.int16),
        ix_dc_inhibition=jnp.array([5], dtype=jnp.int16),
        species_inhibition=jnp.array([4], dtype=jnp.int16),
        species_activation=jnp.array([5], dtype=jnp.int16),
    ),
    IrreversibleMichaelisMenten(  # PROT
        kcat_ix=9,
        enzyme_ix=9,
        km_ix=jnp.array([18], dtype=jnp.int16),
        ki_ix=jnp.array([], dtype=jnp.int16),
        reactant_stoichiometry=jnp.array([-1.0], dtype=jnp.int16),
        ix_substrate=jnp.array([0], dtype=jnp.int16),
        ix_ki_species=jnp.array([], dtype=jnp.int16),
        substrate_km_positions=jnp.array([0], dtype=jnp.int16),
        substrate_reactant_positions=jnp.array([0], dtype=jnp.int16),
    ),
]
steady_state = jnp.array(
    [
        4.233000e-05,  # met-L
        3.099670e-05,  # amet
        2.170170e-07,  # ahcys
        3.521780e-06,  # hcys
        6.534400e-06,  # 5mthf
    ]
)
model = RateEquationModel(parameters, structure, rate_equations)



from jax import numpy as jnp
from jaxtyping import PyTree, Scalar

from enzax.rate_equation import ConcArray, RateEquation


def get_drain_flux(sign: Scalar, log_v: Scalar) -> Scalar:
    """Get the flux of a drain reaction.

    :param sign: a scalar value (should be either one or zero) representing the direction of the reaction.

    :param log_v: a scalar representing the magnitude of the reaction, on log scale.

    """  # Noqa: E501
    return sign * jnp.exp(log_v)


class Drain(RateEquation):
    """A drain reaction."""

    sign: Scalar
    drain_ix: int

    def __call__(self, conc: ConcArray, parameters: PyTree) -> Scalar:
        """Get the flux of a drain reaction."""
        log_v = parameters.log_drain[self.drain_ix]
        return get_drain_flux(self.sign, log_v)



"""Code for MCMC-based Bayesian inference on kinetic models."""

import functools
from typing import Callable, TypedDict, Unpack

import arviz as az
import blackjax
import chex
import jax
from jax._src.random import KeyArray
import jax.numpy as jnp
from jax.scipy.stats import norm
from jaxtyping import Array, Float, PyTree, ScalarLike

from enzax.kinetic_model import (
    RateEquationModel,
    KineticModelStructure,
)
from enzax.parameters import AllostericMichaelisMentenParameterSet
from enzax.rate_equation import RateEquation
from enzax.steady_state import get_kinetic_model_steady_state


@chex.dataclass
class ObservationSet:
    """Measurements from a single experiment."""

    conc: Float[Array, " m"]
    flux: Float[Array, " n"]
    enzyme: Float[Array, " e"]
    conc_scale: ScalarLike
    flux_scale: ScalarLike
    enzyme_scale: ScalarLike


@chex.dataclass
class AllostericMichaelisMentenPriorSet:
    """Priors for an allosteric Michaelis-Menten model."""

    log_kcat: Float[Array, "2 n_enzyme"]
    log_enzyme: Float[Array, "2 n_enzyme"]
    log_drain: Float[Array, "2 n_drain"]
    dgf: Float[Array, "2 n_metabolite"]
    log_km: Float[Array, "2 n_km"]
    log_ki: Float[Array, "2 n_ki"]
    log_conc_unbalanced: Float[Array, "2 n_unbalanced"]
    temperature: Float[Array, "2"]
    log_transfer_constant: Float[Array, "2 n_allosteric_enzyme"]
    log_dissociation_constant: Float[Array, "2 n_allosteric_effect"]


class AdaptationKwargs(TypedDict):
    """Keyword arguments to the blackjax function window_adaptation."""

    initial_step_size: float
    max_num_doublings: int
    is_mass_matrix_diagonal: bool
    target_acceptance_rate: float


def ind_normal_prior_logdensity(param, prior: Float[Array, "2 _"]):
    """Total log density for an independent normal distribution."""
    return norm.logpdf(param, loc=prior[0], scale=prior[1]).sum()


def posterior_logdensity_amm(
    parameters: AllostericMichaelisMentenParameterSet,
    structure: KineticModelStructure,
    rate_equations: list[RateEquation],
    obs: ObservationSet,
    prior: AllostericMichaelisMentenPriorSet,
    guess: Float[Array, " n_balanced"],
):
    """Get the log density for an allosteric Michaelis-Menten model."""
    model = RateEquationModel(parameters, structure, rate_equations)
    steady = get_kinetic_model_steady_state(model, guess)
    flux = model.flux(steady)
    conc = jnp.zeros(model.structure.S.shape[0])
    conc = conc.at[model.structure.balanced_species].set(steady)
    conc = conc.at[model.structure.unbalanced_species].set(
        jnp.exp(parameters.log_conc_unbalanced)
    )
    likelihood_logdensity = (
        norm.logpdf(jnp.log(obs.conc), jnp.log(conc), obs.conc_scale).sum()
        + norm.logpdf(obs.flux, flux[0], obs.flux_scale).sum()
        + norm.logpdf(
            jnp.log(obs.enzyme), parameters.log_enzyme, obs.enzyme_scale
        ).sum()
    )
    prior_logdensity = (
        ind_normal_prior_logdensity(parameters.log_kcat, prior.log_kcat)
        + ind_normal_prior_logdensity(parameters.log_enzyme, prior.log_enzyme)
        + ind_normal_prior_logdensity(parameters.log_drain, prior.log_drain)
        + ind_normal_prior_logdensity(parameters.dgf, prior.dgf)
        + ind_normal_prior_logdensity(parameters.log_km, prior.log_km)
        + ind_normal_prior_logdensity(
            parameters.log_conc_unbalanced, prior.log_conc_unbalanced
        )
        + ind_normal_prior_logdensity(parameters.temperature, prior.temperature)
        + ind_normal_prior_logdensity(parameters.log_ki, prior.log_ki)
        + ind_normal_prior_logdensity(
            parameters.log_transfer_constant, prior.log_transfer_constant
        )
        + ind_normal_prior_logdensity(
            parameters.log_dissociation_constant,
            prior.log_dissociation_constant,
        )
    )
    return prior_logdensity + likelihood_logdensity


@functools.partial(jax.jit, static_argnames=["kernel", "num_samples"])
def _inference_loop(rng_key, kernel, initial_state, num_samples):
    """Run MCMC with blackjax."""

    def one_step(state, rng_key):
        state, info = kernel(rng_key, state)
        return state, (state, info)

    keys = jax.random.split(rng_key, num_samples)
    _, (states, info) = jax.lax.scan(one_step, initial_state, keys)
    return states, info


def run_nuts(
    logdensity_fn: Callable,
    rng_key: KeyArray,
    init_parameters: PyTree,
    num_warmup: int,
    num_samples: int,
    **adapt_kwargs: Unpack[AdaptationKwargs],
):
    """Run the default NUTS algorithm with blackjax."""
    warmup = blackjax.window_adaptation(
        blackjax.nuts,
        logdensity_fn,
        progress_bar=True,
        **adapt_kwargs,
    )
    rng_key, warmup_key = jax.random.split(rng_key)
    (initial_state, tuned_parameters), (_, info, _) = warmup.run(
        warmup_key,
        init_parameters,
        num_steps=num_warmup,  #  type: ignore
    )
    rng_key, sample_key = jax.random.split(rng_key)
    nuts_kernel = blackjax.nuts(logdensity_fn, **tuned_parameters).step
    states, info = _inference_loop(
        sample_key,
        kernel=nuts_kernel,
        initial_state=initial_state,
        num_samples=num_samples,
    )
    return states, info


def ind_prior_from_truth(truth: Float[Array, " _"], sd: ScalarLike):
    """Get a set of independent priors centered at the true parameter values.

    Note that the standard deviation currently has to be the same for
    all parameters.

    """
    return jnp.vstack((truth, jnp.full(truth.shape, sd)))


def get_idata(samples, info, coords=None, dims=None) -> az.InferenceData:
    """Get an arviz InferenceData from a blackjax NUTS output."""
    sample_dict = {
        k: jnp.expand_dims(getattr(samples.position, k), 0)
        for k in samples.position.__dataclass_fields__.keys()
    }
    posterior = az.convert_to_inference_data(
        sample_dict,
        group="posterior",
        coords=coords,
        dims=dims,
    )
    sample_stats = az.convert_to_inference_data(
        {"diverging": info.is_divergent}, group="sample_stats"
    )
    idata = az.concat(posterior, sample_stats)
    assert idata is not None
    return idata



from jax import numpy as jnp
from jaxtyping import Array, Float, Int, PyTree, Scalar

from enzax.rate_equation import ConcArray
from enzax.rate_equations.michaelis_menten import (
    IrreversibleMichaelisMenten,
    MichaelisMenten,
    ReversibleMichaelisMenten,
)


def generalised_mwc_effect(
    conc_inhibitor: Float[Array, " n_inhibition"],
    dc_inhibitor: Float[Array, " n_inhibition"],
    conc_activator: Float[Array, " n_activation"],
    dc_activator: Float[Array, " n_activation"],
    free_enzyme_ratio: Scalar,
    tc: Scalar,
    subunits: int,
) -> Scalar:
    """Get the allosteric effect on a rate.

    The equation is generalised Monod Wyman Changeux model as presented in Popova and Sel'kov 1975: https://doi.org/10.1016/0014-5793(75)80034-2.

    """  # noqa: E501
    qnum = 1 + jnp.sum(conc_inhibitor / dc_inhibitor)
    qdenom = 1 + jnp.sum(conc_activator / dc_activator)
    out = 1.0 / (1 + tc * (free_enzyme_ratio * qnum / qdenom) ** subunits)
    return out


class GeneralisedMWC(MichaelisMenten):
    """Mixin class for allosteric rate laws, assuming generalised MWC kinetics.

    See Popova and Sel'kov 1975 for the rate law: https://doi.org/10.1016/0014-5793(75)80034-2.

    Note that it is assumed there is a free_enzyme_ratio method available - that is why this is a subclass of MichaelisMenten rather than RateEquation.
    """  # noqa: E501

    subunits: int
    tc_ix: int
    ix_dc_activation: Int[Array, " n_activation"]
    ix_dc_inhibition: Int[Array, " n_inhibition"]
    species_activation: Int[Array, " n_activation"]
    species_inhibition: Int[Array, " n_inhibition"]

    def get_tc(self, parameters: PyTree) -> Scalar:
        return jnp.exp(parameters.log_transfer_constant[self.tc_ix])

    def get_dc_activation(self, parameters: PyTree) -> Scalar:
        return jnp.exp(
            parameters.log_dissociation_constant[self.ix_dc_activation],
        )

    def get_dc_inhibition(self, parameters: PyTree) -> Scalar:
        return jnp.exp(
            parameters.log_dissociation_constant[self.ix_dc_inhibition],
        )

    def allosteric_effect(self, conc: ConcArray, parameters: PyTree) -> Scalar:
        return generalised_mwc_effect(
            conc_inhibitor=conc[self.species_inhibition],
            conc_activator=conc[self.species_activation],
            free_enzyme_ratio=self.free_enzyme_ratio(conc, parameters),
            tc=self.get_tc(parameters),
            dc_inhibitor=self.get_dc_inhibition(parameters),
            dc_activator=self.get_dc_activation(parameters),
            subunits=self.subunits,
        )


class AllostericIrreversibleMichaelisMenten(
    GeneralisedMWC, IrreversibleMichaelisMenten
):
    """A reaction with irreversible Michaelis Menten kinetics and allostery."""

    def __call__(self, conc: Float[Array, " n"], parameters: PyTree) -> Scalar:
        """The flux of an irreversible allosteric Michaelis Menten reaction."""
        allosteric_effect = self.allosteric_effect(conc, parameters)
        non_allosteric_rate = super().__call__(conc, parameters)
        return non_allosteric_rate * allosteric_effect


class AllostericReversibleMichaelisMenten(
    GeneralisedMWC,
    ReversibleMichaelisMenten,
):
    """A reaction with reversible Michaelis Menten kinetics and allostery."""

    def __call__(self, conc: ConcArray, parameters: PyTree) -> Scalar:
        """The flux of an allosteric reversible Michaelis Menten reaction."""
        allosteric_effect = self.allosteric_effect(conc, parameters)
        non_allosteric_rate = super().__call__(conc, parameters)
        return non_allosteric_rate * allosteric_effect



"""A simple linear kinetic model."""

from jax import numpy as jnp

from enzax.kinetic_model import (
    RateEquationModel,
    KineticModelStructure,
)
from enzax.rate_equations import (
    AllostericReversibleMichaelisMenten,
    ReversibleMichaelisMenten,
)
from enzax.parameters import AllostericMichaelisMentenParameterSet

parameters = AllostericMichaelisMentenParameterSet(
    log_kcat=jnp.array([-0.1, 0.0, 0.1]),
    log_enzyme=jnp.log(jnp.array([0.3, 0.2, 0.1])),
    dgf=jnp.array([-3, -1.0]),
    log_km=jnp.array([0.1, -0.2, 0.5, 0.0, -1.0, 0.5]),
    log_ki=jnp.array([1.0]),
    log_conc_unbalanced=jnp.log(jnp.array([0.5, 0.1])),
    temperature=jnp.array(310.0),
    log_transfer_constant=jnp.array([-0.2, 0.3]),
    log_dissociation_constant=jnp.array([-0.1, 0.2]),
    log_drain=jnp.array([]),
)
structure = KineticModelStructure(
    S=jnp.array(
        [[-1, 0, 0], [1, -1, 0], [0, 1, -1], [0, 0, 1]], dtype=jnp.float64
    ),
    balanced_species=jnp.array([1, 2]),
    unbalanced_species=jnp.array([0, 3]),
)
rate_equations = [
    AllostericReversibleMichaelisMenten(
        kcat_ix=0,
        enzyme_ix=0,
        km_ix=jnp.array([0, 1], dtype=jnp.int16),
        ki_ix=jnp.array([], dtype=jnp.int16),
        reactant_stoichiometry=jnp.array([-1, 1], dtype=jnp.int16),
        reactant_to_dgf=jnp.array([0, 0], dtype=jnp.int16),
        ix_ki_species=jnp.array([], dtype=jnp.int16),
        substrate_km_positions=jnp.array([0], dtype=jnp.int16),
        substrate_reactant_positions=jnp.array([0], dtype=jnp.int16),
        ix_substrate=jnp.array([0], dtype=jnp.int16),
        ix_product=jnp.array([1], dtype=jnp.int16),
        ix_reactants=jnp.array([0, 1], dtype=jnp.int16),
        product_reactant_positions=jnp.array([1], dtype=jnp.int16),
        product_km_positions=jnp.array([1], dtype=jnp.int16),
        water_stoichiometry=jnp.array(0.0),
        tc_ix=0,
        ix_dc_inhibition=jnp.array([], dtype=jnp.int16),
        ix_dc_activation=jnp.array([0], dtype=jnp.int16),
        species_activation=jnp.array([2], dtype=jnp.int16),
        species_inhibition=jnp.array([], dtype=jnp.int16),
        subunits=1,
    ),
    AllostericReversibleMichaelisMenten(
        kcat_ix=1,
        enzyme_ix=1,
        km_ix=jnp.array([2, 3], dtype=jnp.int16),
        ki_ix=jnp.array([0]),
        reactant_stoichiometry=jnp.array([-1, 1], dtype=jnp.int16),
        reactant_to_dgf=jnp.array([0, 1], dtype=jnp.int16),
        ix_ki_species=jnp.array([1]),
        substrate_km_positions=jnp.array([0], dtype=jnp.int16),
        substrate_reactant_positions=jnp.array([0], dtype=jnp.int16),
        ix_substrate=jnp.array([1], dtype=jnp.int16),
        ix_product=jnp.array([2], dtype=jnp.int16),
        ix_reactants=jnp.array([1, 2], dtype=jnp.int16),
        product_reactant_positions=jnp.array([1], dtype=jnp.int16),
        product_km_positions=jnp.array([1], dtype=jnp.int16),
        water_stoichiometry=jnp.array(0.0),
        tc_ix=1,
        ix_dc_inhibition=jnp.array([1], dtype=jnp.int16),
        ix_dc_activation=jnp.array([], dtype=jnp.int16),
        species_activation=jnp.array([], dtype=jnp.int16),
        species_inhibition=jnp.array([1], dtype=jnp.int16),
        subunits=1,
    ),
    ReversibleMichaelisMenten(
        kcat_ix=2,
        enzyme_ix=2,
        km_ix=jnp.array([4, 5], dtype=jnp.int16),
        ki_ix=jnp.array([], dtype=jnp.int16),
        ix_substrate=jnp.array([2], dtype=jnp.int16),
        ix_product=jnp.array([3], dtype=jnp.int16),
        ix_reactants=jnp.array([2, 3], dtype=jnp.int16),
        reactant_to_dgf=jnp.array([1, 1], dtype=jnp.int16),
        reactant_stoichiometry=jnp.array([-1, 1], dtype=jnp.int16),
        ix_ki_species=jnp.array([], dtype=jnp.int16),
        substrate_km_positions=jnp.array([0], dtype=jnp.int16),
        substrate_reactant_positions=jnp.array([0], dtype=jnp.int16),
        product_reactant_positions=jnp.array([1], dtype=jnp.int16),
        product_km_positions=jnp.array([1], dtype=jnp.int16),
        water_stoichiometry=jnp.array(0.0),
    ),
]
steady_state = jnp.array([0.43658744, 0.12695706])
model = RateEquationModel(parameters, structure, rate_equations)



"""Module with parameters and parameter sets.

These are not required, but they do provide handy shape checking, for example to ensure that your parameter set has the same number of enzymes and kcat parameters.

"""  # Noqa: E501

import equinox as eqx
from jaxtyping import Array, Float, Scalar, jaxtyped
from typeguard import typechecked

LogKcat = Float[Array, " n_enzyme"]
LogEnzyme = Float[Array, " n_enzyme"]
Dgf = Float[Array, " n_metabolite"]
LogKm = Float[Array, " n_km"]
LogKi = Float[Array, " n_ki"]
LogConcUnbalanced = Float[Array, " n_unbalanced"]
LogDrain = Float[Array, " n_drain"]
LogTransferConstant = Float[Array, " n_allosteric_enzyme"]
LogDissociationConstant = Float[Array, " n_allosteric_effect"]


@jaxtyped(typechecker=typechecked)
class MichaelisMentenParameterSet(eqx.Module):
    """Parameters for a model with Michaelis Menten kinetics.

    This kind of parameter set supports models with the following rate laws:

      - enzax.rate_equations.drain.Drain
      - enzax.rate_equations.michaelis_menten.IrreversibleMichaelisMenten
      - enzax.rate_equations.michaelis_menten.ReversibleMichaelisMenten

    """

    log_kcat: LogKcat
    log_enzyme: LogEnzyme
    dgf: Dgf
    log_km: LogKm
    log_ki: LogKi
    log_conc_unbalanced: LogConcUnbalanced
    temperature: Scalar
    log_drain: LogDrain


class AllostericMichaelisMentenParameterSet(MichaelisMentenParameterSet):
    """Parameters for a model with Michaelis Menten kinetics, with allostery.

    Reactions can be any out of:

      - drain.Drain
      - michaelis_menten.IrreversibleMichaelisMenten
      - michaelis_menten.ReversibleMichaelisMenten
      - generalised_mwc.AllostericIrreversibleMichaelisMenten
      - generalised_mwc.AllostericReversibleMichaelisMenten

    """

    log_transfer_constant: LogTransferConstant
    log_dissociation_constant: LogDissociationConstant



"""Demonstration of how to make a Bayesian kinetic model with enzax."""

import functools
import logging
import warnings

import arviz as az
import jax
from jax import numpy as jnp

from enzax.examples import methionine
from enzax.mcmc import (
    ObservationSet,
    AllostericMichaelisMentenPriorSet,
    get_idata,
    ind_prior_from_truth,
    posterior_logdensity_amm,
    run_nuts,
)
from enzax.steady_state import get_kinetic_model_steady_state

SEED = 1234

jax.config.update("jax_enable_x64", True)


def main():
    """Demonstrate How to make a Bayesian kinetic model with enzax."""
    structure = methionine.structure
    rate_equations = methionine.rate_equations
    true_parameters = methionine.parameters
    true_model = methionine.model
    default_state_guess = jnp.full((5,), 0.01)
    true_states = get_kinetic_model_steady_state(
        true_model, default_state_guess
    )
    prior = AllostericMichaelisMentenPriorSet(
        log_kcat=ind_prior_from_truth(true_parameters.log_kcat, 0.1),
        log_enzyme=ind_prior_from_truth(true_parameters.log_enzyme, 0.1),
        log_drain=ind_prior_from_truth(true_parameters.log_drain, 0.1),
        dgf=ind_prior_from_truth(true_parameters.dgf, 0.1),
        log_km=ind_prior_from_truth(true_parameters.log_km, 0.1),
        log_conc_unbalanced=ind_prior_from_truth(
            true_parameters.log_conc_unbalanced, 0.1
        ),
        temperature=ind_prior_from_truth(true_parameters.temperature, 0.1),
        log_ki=ind_prior_from_truth(true_parameters.log_ki, 0.1),
        log_transfer_constant=ind_prior_from_truth(
            true_parameters.log_transfer_constant, 0.1
        ),
        log_dissociation_constant=ind_prior_from_truth(
            true_parameters.log_dissociation_constant, 0.1
        ),
    )
    # get true concentration
    true_conc = jnp.zeros(methionine.structure.S.shape[0])
    true_conc = true_conc.at[methionine.structure.balanced_species].set(
        true_states
    )
    true_conc = true_conc.at[methionine.structure.unbalanced_species].set(
        jnp.exp(true_parameters.log_conc_unbalanced)
    )
    # get true flux
    true_flux = true_model.flux(true_states)
    # simulate observations
    error_conc = 0.03
    error_flux = 0.05
    error_enzyme = 0.03
    key = jax.random.key(SEED)
    obs_conc = jnp.exp(jnp.log(true_conc) + jax.random.normal(key) * error_conc)
    obs_enzyme = jnp.exp(
        true_parameters.log_enzyme + jax.random.normal(key) * error_enzyme
    )
    obs_flux = true_flux + jax.random.normal(key) * error_conc
    obs = ObservationSet(
        conc=obs_conc,
        flux=obs_flux,
        enzyme=obs_enzyme,
        conc_scale=error_conc,
        flux_scale=error_flux,
        enzyme_scale=error_enzyme,
    )
    pldf = functools.partial(
        posterior_logdensity_amm,
        obs=obs,
        prior=prior,
        structure=structure,
        rate_equations=rate_equations,
        guess=default_state_guess,
    )
    samples, info = run_nuts(
        pldf,
        key,
        true_parameters,
        num_warmup=200,
        num_samples=200,
        initial_step_size=0.0001,
        max_num_doublings=10,
        is_mass_matrix_diagonal=False,
        target_acceptance_rate=0.95,
    )
    idata = get_idata(
        samples, info, coords=methionine.coords, dims=methionine.dims
    )
    print(az.summary(idata))
    if jnp.any(info.is_divergent):
        n_divergent = info.is_divergent.sum()
        msg = f"There were {n_divergent} post-warmup divergent transitions."
        warnings.warn(msg)
    else:
        logging.info("No post-warmup divergent transitions!")
    print("True parameter values vs posterior:")
    for param in true_parameters.__dataclass_fields__.keys():
        true_val = getattr(true_parameters, param)
        model_low = jnp.quantile(getattr(samples.position, param), 0.01, axis=0)
        model_high = jnp.quantile(
            getattr(samples.position, param), 0.99, axis=0
        )
        print(f" {param}:")
        print(f"  true value: {true_val}")
        print(f"  posterior 1%: {model_low}")
        print(f"  posterior 99%: {model_high}")


if __name__ == "__main__":
    main()



"""Unit tests for rate equations."""

from jax import numpy as jnp
from enzax.rate_equations import (
    AllostericIrreversibleMichaelisMenten,
    AllostericReversibleMichaelisMenten,
    IrreversibleMichaelisMenten,
    ReversibleMichaelisMenten,
)
from enzax.parameters import AllostericMichaelisMentenParameterSet

EXAMPLE_CONC = jnp.array([0.5, 0.2, 0.1, 0.3])
EXAMPLE_PARAMETERS = AllostericMichaelisMentenParameterSet(
    log_kcat=jnp.array([-0.1]),
    log_enzyme=jnp.log(jnp.array([0.3])),
    dgf=jnp.array([-3, -1.0]),
    log_km=jnp.array([0.1, -0.2]),
    log_ki=jnp.array([1.0]),
    log_conc_unbalanced=jnp.log(jnp.array([0.5, 0.3])),
    temperature=jnp.array(310.0),
    log_transfer_constant=jnp.array([-0.2, 0.3]),
    log_dissociation_constant=jnp.array([-0.1, 0.2]),
    log_drain=jnp.array([]),
)


def test_irreversible_michaelis_menten():
    expected_rate = 0.08455524
    f = IrreversibleMichaelisMenten(
        kcat_ix=0,
        enzyme_ix=0,
        km_ix=jnp.array([0, 1], dtype=jnp.int16),
        ki_ix=jnp.array([], dtype=jnp.int16),
        reactant_stoichiometry=jnp.array([-1, 1], dtype=jnp.int16),
        ix_ki_species=jnp.array([], dtype=jnp.int16),
        substrate_km_positions=jnp.array([0], dtype=jnp.int16),
        substrate_reactant_positions=jnp.array([0], dtype=jnp.int16),
        ix_substrate=jnp.array([0], dtype=jnp.int16),
    )
    rate = f(EXAMPLE_CONC, EXAMPLE_PARAMETERS)
    assert jnp.isclose(rate, expected_rate)


def test_reversible_michaelis_menten():
    expected_rate = 0.04342889
    f = ReversibleMichaelisMenten(
        kcat_ix=0,
        enzyme_ix=0,
        km_ix=jnp.array([0, 1], dtype=jnp.int16),
        ki_ix=jnp.array([], dtype=jnp.int16),
        reactant_stoichiometry=jnp.array([-1, 1], dtype=jnp.int16),
        reactant_to_dgf=jnp.array([0, 0], dtype=jnp.int16),
        ix_ki_species=jnp.array([], dtype=jnp.int16),
        substrate_km_positions=jnp.array([0], dtype=jnp.int16),
        substrate_reactant_positions=jnp.array([0], dtype=jnp.int16),
        ix_substrate=jnp.array([0], dtype=jnp.int16),
        ix_product=jnp.array([1], dtype=jnp.int16),
        ix_reactants=jnp.array([0, 1], dtype=jnp.int16),
        product_reactant_positions=jnp.array([1], dtype=jnp.int16),
        product_km_positions=jnp.array([1], dtype=jnp.int16),
        water_stoichiometry=jnp.array(0.0),
    )
    rate = f(EXAMPLE_CONC, EXAMPLE_PARAMETERS)
    assert jnp.isclose(rate, expected_rate)


def test_allosteric_irreversible_michaelis_menten():
    expected_rate = 0.05608589
    f = AllostericIrreversibleMichaelisMenten(
        kcat_ix=0,
        enzyme_ix=0,
        km_ix=jnp.array([0, 1], dtype=jnp.int16),
        ki_ix=jnp.array([], dtype=jnp.int16),
        reactant_stoichiometry=jnp.array([-1, 1], dtype=jnp.int16),
        ix_ki_species=jnp.array([], dtype=jnp.int16),
        substrate_km_positions=jnp.array([0], dtype=jnp.int16),
        substrate_reactant_positions=jnp.array([0], dtype=jnp.int16),
        ix_substrate=jnp.array([0], dtype=jnp.int16),
        tc_ix=0,
        ix_dc_inhibition=jnp.array([], dtype=jnp.int16),
        ix_dc_activation=jnp.array([0], dtype=jnp.int16),
        species_activation=jnp.array([2], dtype=jnp.int16),
        species_inhibition=jnp.array([], dtype=jnp.int16),
        subunits=1,
    )
    rate = f(EXAMPLE_CONC, EXAMPLE_PARAMETERS)
    assert jnp.isclose(rate, expected_rate)


def test_allosteric_reversible_michaelis_menten():
    expected_rate = 0.03027414
    f = AllostericReversibleMichaelisMenten(
        kcat_ix=0,
        enzyme_ix=0,
        km_ix=jnp.array([0, 1], dtype=jnp.int16),
        ki_ix=jnp.array([], dtype=jnp.int16),
        reactant_stoichiometry=jnp.array([-1, 1], dtype=jnp.int16),
        reactant_to_dgf=jnp.array([0, 0], dtype=jnp.int16),
        ix_ki_species=jnp.array([], dtype=jnp.int16),
        substrate_km_positions=jnp.array([0], dtype=jnp.int16),
        substrate_reactant_positions=jnp.array([0], dtype=jnp.int16),
        ix_substrate=jnp.array([0], dtype=jnp.int16),
        ix_product=jnp.array([1], dtype=jnp.int16),
        ix_reactants=jnp.array([0, 1], dtype=jnp.int16),
        product_reactant_positions=jnp.array([1], dtype=jnp.int16),
        product_km_positions=jnp.array([1], dtype=jnp.int16),
        water_stoichiometry=jnp.array(0.0),
        tc_ix=0,
        ix_dc_inhibition=jnp.array([], dtype=jnp.int16),
        ix_dc_activation=jnp.array([0], dtype=jnp.int16),
        species_activation=jnp.array([2], dtype=jnp.int16),
        species_inhibition=jnp.array([], dtype=jnp.int16),
        subunits=1,
    )
    rate = f(EXAMPLE_CONC, EXAMPLE_PARAMETERS)
    assert jnp.isclose(rate, expected_rate)



from abc import abstractmethod
from jax import numpy as jnp
from jaxtyping import Array, Float, Int, PyTree, Scalar

from enzax.rate_equation import RateEquation, ConcArray


def numerator_mm(
    conc: ConcArray,
    km: Float[Array, " n"],
    ix_substrate: Int[Array, " n_substrate"],
    substrate_km_positions: Int[Array, " n_substrate"],
) -> Scalar:
    """Get the product of each substrate's concentration over its km.

    This quantity is the numerator in a Michaelis Menten reaction's rate equation
    """  # Noqa: E501
    return jnp.prod((conc[ix_substrate] / km[substrate_km_positions]))


class MichaelisMenten(RateEquation):
    """Base class for Michaelis Menten rate equations.

    Subclasses need to implement the __call__ and free_enzyme_ratio methods.

    """

    kcat_ix: int
    enzyme_ix: int
    km_ix: Int[Array, " n"]
    ki_ix: Int[Array, " n_ki"]
    reactant_stoichiometry: Float[Array, " n"]
    ix_substrate: Int[Array, " n_substrate"]
    ix_ki_species: Int[Array, " n_ki"]
    substrate_km_positions: Int[Array, " n_substrate"]
    substrate_reactant_positions: Int[Array, " n_substrate"]

    def get_kcat(self, parameters: PyTree) -> Scalar:
        return jnp.exp(parameters.log_kcat[self.kcat_ix])

    def get_km(self, parameters: PyTree) -> Scalar:
        return jnp.exp(parameters.log_km[self.km_ix])

    def get_ki(self, parameters: PyTree) -> Scalar:
        return jnp.exp(parameters.log_ki[self.ki_ix])

    def get_enzyme(self, parameters: PyTree) -> Scalar:
        return jnp.exp(parameters.log_enzyme[self.enzyme_ix])

    @abstractmethod
    def free_enzyme_ratio(
        self,
        conc: ConcArray,
        parameters: PyTree,
    ) -> Scalar: ...


def free_enzyme_ratio_imm(
    conc_sub: Float[Array, " n_substrate"],
    km_sub: Float[Array, " n_substrate"],
    stoich_sub: Float[Array, " n_substrate"],
    ki: Float[Array, " n_ki"],
    conc_inhibitor: Float[Array, " n_ki"],
) -> Scalar:
    """Free enzyme ratio for irreversible Michaelis Menten reactions."""
    return 1.0 / (
        jnp.prod(((conc_sub / km_sub) + 1) ** jnp.abs(stoich_sub))
        + jnp.sum(conc_inhibitor / ki)
    )


class IrreversibleMichaelisMenten(MichaelisMenten):
    """A reaction with irreversible Michaelis Menten kinetics."""

    def free_enzyme_ratio(self, conc, parameters):
        return free_enzyme_ratio_imm(
            conc_sub=conc[self.ix_substrate],
            km_sub=self.get_km(parameters)[self.substrate_km_positions],
            ki=self.get_ki(parameters),
            conc_inhibitor=conc[self.ix_ki_species],
            stoich_sub=self.reactant_stoichiometry[
                self.substrate_reactant_positions
            ],
        )

    def __call__(self, conc: Float[Array, " n"], parameters: PyTree) -> Scalar:
        """Get flux of a reaction with irreversible Michaelis Menten kinetics."""  # noqa: E501
        kcat = self.get_kcat(parameters)
        enzyme = self.get_enzyme(parameters)
        km = self.get_km(parameters)
        numerator = numerator_mm(
            conc=conc,
            km=km,
            ix_substrate=self.ix_substrate,
            substrate_km_positions=self.substrate_km_positions,
        )
        free_enzyme_ratio = self.free_enzyme_ratio(conc, parameters)
        return kcat * enzyme * numerator * free_enzyme_ratio


def get_reversibility(
    conc: Float[Array, " n"],
    water_stoichiometry: Scalar,
    dgf: Float[Array, " n_reactant"],
    temperature: Scalar,
    reactant_stoichiometry: Float[Array, " n_reactant"],
    ix_reactants: Int[Array, " n_reactant"],
) -> Scalar:
    """Get the reversibility of a reaction.

    Hard coded water dgf is taken from <http://equilibrator.weizmann.ac.il/metabolite?compoundId=C00001>.

    """
    RT = temperature * 0.008314
    dgf_water = -150.9
    dgr = reactant_stoichiometry @ dgf + water_stoichiometry * dgf_water
    quotient = reactant_stoichiometry @ jnp.log(conc[ix_reactants])
    out = 1.0 - jnp.exp(((dgr + RT * quotient) / RT))
    return out


def free_enzyme_ratio_rmm(
    conc_sub: Float[Array, " n_substrate"],
    km_sub: Float[Array, " n_substrate"],
    stoich_sub: Float[Array, " n_substrate"],
    conc_prod: Float[Array, " n_product"],
    km_prod: Float[Array, " n_prod"],
    stoich_prod: Float[Array, " n_prod"],
    conc_inhibitor: Float[Array, " n_ki"],
    ki: Float[Array, " n_ki"],
) -> Scalar:
    """The free enzyme ratio for a reversible Michaelis Menten reaction."""
    return 1.0 / (
        -1.0
        + jnp.prod(((conc_sub / km_sub) + 1.0) ** jnp.abs(stoich_sub))
        + jnp.prod(((conc_prod / km_prod) + 1.0) ** jnp.abs(stoich_prod))
        + jnp.sum(conc_inhibitor / ki)
    )


class ReversibleMichaelisMenten(MichaelisMenten):
    """A reaction with reversible Michaelis Menten kinetics."""

    ix_product: Int[Array, " n_product"]
    ix_reactants: Int[Array, " n_reactant"]
    product_reactant_positions: Int[Array, " n_product"]
    product_km_positions: Int[Array, " n_product"]
    water_stoichiometry: Scalar
    reactant_to_dgf: Int[Array, " n_reactant"]

    def _get_dgf(self, parameters: PyTree):
        return parameters.dgf[self.reactant_to_dgf]

    def free_enzyme_ratio(self, conc: ConcArray, parameters: PyTree) -> Scalar:
        return free_enzyme_ratio_rmm(
            conc_sub=conc[self.ix_substrate],
            km_sub=self.get_km(parameters)[self.substrate_reactant_positions],
            stoich_sub=self.reactant_stoichiometry[
                self.substrate_reactant_positions
            ],
            conc_prod=conc[self.ix_product],
            km_prod=self.get_km(parameters)[self.product_reactant_positions],
            stoich_prod=self.reactant_stoichiometry[
                self.product_reactant_positions
            ],
            conc_inhibitor=conc[self.ix_ki_species],
            ki=self.get_ki(parameters),
        )

    def __call__(self, conc: Float[Array, " n"], parameters: PyTree) -> Scalar:
        """Get flux of a reaction with reversible Michaelis Menten kinetics.

        :param conc: A 1D array of non-negative numbers representing concentrations of the species that the reaction produces and consumes.

        """  # noqa: E501
        kcat = self.get_kcat(parameters)
        enzyme = self.get_enzyme(parameters)
        reversibility = get_reversibility(
            conc=conc,
            water_stoichiometry=self.water_stoichiometry,
            dgf=self._get_dgf(parameters),
            temperature=parameters.temperature,
            reactant_stoichiometry=self.reactant_stoichiometry,
            ix_reactants=self.ix_reactants,
        )
        numerator = numerator_mm(
            conc=conc,
            km=self.get_km(parameters),
            ix_substrate=self.ix_substrate,
            substrate_km_positions=self.substrate_km_positions,
        )
        free_enzyme_ratio = self.free_enzyme_ratio(conc, parameters)
        return reversibility * kcat * enzyme * numerator * free_enzyme_ratio



"""Module containing rate equations for enzyme-catalysed reactions."""

from abc import ABC, abstractmethod
from equinox import Module

from jaxtyping import Array, Float, PyTree, Scalar


ConcArray = Float[Array, " n"]


class RateEquation(Module, ABC):
    """Abstract definition of a rate equation.

    A rate equation is an equinox [Module](https://docs.kidger.site/equinox/api/module/module/) with a `__call__` method that takes in a 1 dimensional array of concentrations and an arbitrary PyTree of parameters, returning a scalar value representing a single flux.
    """  # Noqa: E501

    @abstractmethod
    def __call__(self, conc: ConcArray, parameters: PyTree) -> Scalar: ...



"""Module containing enzax's definition of a kinetic model."""

from abc import ABC, abstractmethod

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Float, Int, PyTree, ScalarLike, jaxtyped
from typeguard import typechecked

from enzax.rate_equation import RateEquation


@jaxtyped(typechecker=typechecked)
class KineticModelStructure(eqx.Module):
    """Structural information about a kinetic model."""

    S: Float[Array, " s r"]
    balanced_species: Int[Array, " n_balanced"]
    unbalanced_species: Int[Array, " n_unbalanced"]


class KineticModel(eqx.Module, ABC):
    """Abstract base class for kinetic models."""

    parameters: PyTree
    structure: KineticModelStructure
    rate_equations: list[RateEquation] = eqx.field(default_factory=list)

    @abstractmethod
    def flux(
        self,
        conc_balanced: Float[Array, " n_balanced"],
    ) -> Float[Array, " n"]: ...

    def dcdt(
        self, t: ScalarLike, conc: Float[Array, " n_balanced"], args=None
    ) -> Float[Array, " n_balanced"]:
        """Get the rate of change of balanced species concentrations.

        Note that the signature is as required for a Diffrax vector field function, hence the redundant variable t and the weird name "args".

        :param t: redundant variable representing time.

        :param conc: a one dimensional array of positive floats representing concentrations of balanced species. Must have same size as self.structure.ix_balanced

        """  # Noqa: E501
        sv = self.structure.S @ self.flux(conc)
        return sv[self.structure.balanced_species]


class RateEquationModel(KineticModel):
    """A kinetic model that specifies its fluxes using RateEquation objects."""

    rate_equations: list[RateEquation] = eqx.field(default_factory=list)

    def flux(
        self,
        conc_balanced: Float[Array, " n_balanced"],
    ) -> Float[Array, " n"]:
        """Get fluxes from balanced species concentrations.

        :param conc_balanced: a one dimensional array of positive floats representing concentrations of balanced species. Must have same size as self.structure.ix_balanced

        :return: a one dimensional array of (possibly negative) floats representing reaction fluxes. Has same size as number of columns of self.structure.S.

        """  # Noqa: E501
        conc = jnp.zeros(self.structure.S.shape[0])
        conc = conc.at[self.structure.balanced_species].set(conc_balanced)
        conc = conc.at[self.structure.unbalanced_species].set(
            jnp.exp(self.parameters.log_conc_unbalanced)
        )
        t = [f(conc, self.parameters) for f in self.rate_equations]
        out = jnp.array(t)
        return out
